//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <vcl\registry.hpp>
#include <filectrl.hpp>
#pragma hdrstop

#include "SetupDlg.h"
#include "WaitDlg.h"
#include "FileMan.h"
#include "HeadMan.h"
#include "MsgWins.h"
#include "DirBrow.h"
#include "setupver.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma link "explbtn"
#pragma resource "*.dfm"
TSetupForm *SetupForm;
//---------------------------------------------------------------------------
__fastcall TSetupForm::TSetupForm(TComponent* Owner)
  : TForm(Owner)
{
   setuplog = new TStringList();
   logging = keeplog;
   log("begin setup");
   log("initializing setup wizard");
   const AnsiString MenuString = "About Setup...";
   MENUITEMINFO info = {  sizeof(MENUITEMINFO),
                         MIIM_TYPE | MIIM_STATE | MIIM_ID ,
                         MFT_STRING,                 // menu type is string
                         MFS_ENABLED,                // menu item is enabled
                         IDS_CUSTOMSYSMENUITEM,      // id of menu item
                         NULL,                       // no sub menu
                         NULL,                       // no check bitmap
                         NULL,                       // no uncheck bitmap
                         0,                          // no item data
                         MenuString.c_str(),         // string of menu item
                         0};                         // last item not used
  MENUITEMINFO sepr = {  sizeof(MENUITEMINFO),
                         MIIM_TYPE | MIIM_STATE | MIIM_ID ,
                         MFT_SEPARATOR,                 // menu type is string
                         MFS_ENABLED,                // menu item is enabled
                         0,  // id of menu item
                         NULL,                       // no sub menu
                         NULL,                       // no check bitmap
                         NULL,                       // no uncheck bitmap
                         0,                          // no item data
                         NULL,                       // string of menu item
                         0};                         // last item not used
   // obtain the menu handle for the system menu
   HMENU SysMenu = GetSystemMenu(Handle, FALSE);
   // Insert the new menu item just after the close menu item. If the TRUE
   InsertMenuItem (SysMenu, SC_CLOSE, true, &sepr);
   InsertMenuItem (SysMenu, 5, false, &info);

   isUpdate = false;
   TempDir = GetCurrentDir();
   AnsiString cap = CopyRight;
   int pos = cap.Pos("\n"); if (pos) { cap.Insert(",  ",pos); cap.Delete(pos+3,1); }
   Copyright->Caption = cap;
 //appexe,appini,appdat need to refer to qs.exe rather than setup.exe
 //they do not know appdir yet
   AppEXE = AppExe; // no directory--only filename
   AppINI = ChangeFileExt(AppEXE,IniExt);
   AppDAT = ChangeFileExt(AppEXE,DatExt);
   releasever = GetAppIni("Version","",AppINI);
   Appname->Caption = AppName+" "+releasever;
   Notebook1->PageIndex = 0;
   ReadRegistry();
   ReadDat();
   ActiveControl = NextBtn;
   OrgAppIcon = Application->Icon;
   Application->Icon = Icon;
   logdir = "";
 }
//---------------------------------------------------------------------------
void __fastcall TSetupForm::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   // only cancelbtn gets us here--mrCancel whether cancel or exit
   if (Notebook1->PageIndex>0 && CancelBtn->Caption!="Exit")
   {
      if (MsgBoxHwnd(Handle,"Quit without installing QuickSilver?",
         AppName,MB_YESNO)==IDNO)
      {
         CanClose = false;
         return;
      }
   }
   SetCurrentDir(AppDIR->Text);
   if (Notebook1->ActivePage=="Done" && CancelBtn->Caption=="Exit" && RunQuickBox->Checked)
      Exec95(AddBackslash(AppDIR->Text)+AppEXE);
   Application->Icon = OrgAppIcon;
   delete setuplog;
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::WMSysCommand(TWMSysCommand &Msg)
{
  if(Msg.CmdType==IDS_CUSTOMSYSMENUITEM)
  {
     MsgBoxHwnd(Handle,SetupName+"  "+ReleaseVer+"\n"+CopyRight,"About...",MB_OK);
     Msg.Result = true;
  }
  else
     TForm::Dispatch(&Msg);
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::log(AnsiString &event)
{
   if (logging && (logdir.IsEmpty() || DirectoryExists(logdir)))
   {
      setuplog->Add(event.LowerCase());
      setuplog->SaveToFile(logdir+SetupLog);
   }
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::ReadRegistry()
{
   log("searching windows registry for existing QS");
   TRegistry *reg = new TRegistry();
   if (reg->OpenKey(QuickKey,false))
   {
      AnsiString appdir;
      GroupBox->Checked = !reg->ReadString("Program Group").IsEmpty();
      if ((appdir=reg->ReadString("AppDir"))!="") AppDIR->Text = appdir;
      reg->CloseKey();

      logdir = AddBackslash(appdir).LowerCase();
      log("found QS "+reg->ReadString("Version")+" in "+logdir.LowerCase());
      if (GroupBox->Checked) GroupBox->Enabled = false;
      if (ShortBox->Checked) ShortBox->Enabled = false;
      isUpdate = true;
   }
   else log("no previous QS installed");

   log("searching windows registry for existing mix2.9");
   if (reg->OpenKey(MixKey,false))
   {
      MixDIR = reg->ReadString("MixDir").LowerCase();
      log("found mix2.9 in "+MixDIR);
      reg->CloseKey();
   }
   else log("no previous mix2.9 installed");

   delete reg;
   if (AppDIR->Text.IsEmpty())
       AppDIR->Text = "C:\\Program Files\\"+AppName;

   if (isUpdate) log("qs directory is "+AppDIR->Text);
   return true;
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::ReadDat()
{
   if (isUpdate)
      log("locating previous QS database");
   AnsiString appdat = AddBackslash(AppDIR->Text)+AppDAT.LowerCase();
   if (!FileExists(appdat)) return;

   TStringList *msg = new TStringList();
   THeaderManager *headers = new THeaderManager(msg);

   GetAppData(msg,DefaultMessage,appdat);

   EAddr->Text = headers->GetDat("From: ",msg);
   Host->Text = headers->GetDat("Host: ",msg);
   delete headers;
   delete msg;
   log("found database "+appdat);
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::WriteRegistry()
{
   bool ok = false;
   TRegistry *reg = new TRegistry();

   try {
      if (reg->OpenKey(QuickKey,true))
      {
         log("writing QS home directory to registry");
         reg->WriteString("AppDir",RemoveBackslash(AppDIR->Text));
         log("writing QS version to registry");
         reg->WriteString("Version",releasever);
         if (GroupBox->Checked) reg->WriteString("Program Group",AppName);
         if (!isUpdate) // don't reposition window if this is not a new install
         {
            int h = Screen->Height>480 ? 500:400;
            int w = Screen->Width>640 ? 750:600;
            int x = (Screen->Width-w)/2;
            int y = (Screen->Height-h)/2;
            reg->WriteInteger("WindowState",wsNormal);
            reg->WriteString("WindowPos",s(x)+","+s(y)+","+s(w)+","+s(h));
         }
         reg->WriteBool("Update",true); // tells QS an update was just installed
         reg->CloseKey();
      }
      ok = true;
   }
   catch (...) {
      MsgBoxHwnd(Handle,
         AnsiString("Unable to write to windows registry!\n\n")+
         AnsiString("Entries NOT saved!"),
        "Registry Error",MB_OK);

      log("writing to the registry failed");
      delete reg;
   }
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::WriteDat()
{
   AnsiString appdat=AddBackslash(AppDIR->Text)+AppDAT;
   if (FileExists(ChangeFileExt(appdat,PgpExt)) && !FileExists(appdat))
      return true;

   TStringList *msg = new TStringList();
   msg->Add("Fcc: outbox");
   msg->Add("Host: "+Host->Text);
   msg->Add("From: "+EAddr->Text);
   msg->Add("Chain: *,*,*,*; copies=1");
   msg->Add("To: ");
   msg->Add("Subject: ");
   msg->Add("");
   msg->Add(CutMarks);
   log("writing default message to database");
   if (PutAppData(DefaultMessage,msg,appdat)!=TOOL_OK)
   {
      MsgBoxHwnd(Handle,"Error storing default message!","Setup Error",MB_OK);
      log("writing default message failed");
      return false;
   }
   delete msg;
   TStringList *art = new TStringList();
   art->Add("Fcc: outbox");
   art->Add("Host: "+Host->Text);
   art->Add("From: "+EAddr->Text);
   art->Add("Chain: *,*,*,*; copies=1");
   art->Add("To: mail2news@anon.lcs.mit.edu");
   art->Add("X-No-Archive: Yes");
   art->Add("Newsgroups: ");
   art->Add("Subject: ");
   art->Add("");
   art->Add("~~");
   log("writing default article to database");
   if (PutAppData(DefaultArticle,art,appdat)!=TOOL_OK)
   {
      MsgBoxHwnd(Handle,"Error storing default article!","Setup Error",MB_OK);
      log("writing default article failed");
      return false;
   }
   delete art;
   return true;
}
//---------------------------------------------------------------------------
AnsiString __fastcall ShortPath(const AnsiString &LongPath)
{
   char *buf = NULL;
   int size = GetShortPathName(LongPath.c_str(),(LPSTR)buf,1);
   buf = new char[size+1];
   GetShortPathName(LongPath.c_str(),(LPSTR)buf,size);
   AnsiString shortpath(buf);
   delete buf;
   return RemoveBackslash(shortpath.LowerCase());
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::CopyFiles()
{
   // are we _NOT_ in a temp directory...(already installed--HERE)
   if (ShortPath(TempDir)==ShortPath(AppDIR->Text)) return true;

   log("creating application directory "+AppDIR->Text);
   if (!DirectoryExists(AppDIR->Text)) ForceDirectories(AppDIR->Text);
   if (!DirectoryExists(AppDIR->Text))
   {
      MsgBoxHwnd(Handle,
         AnsiString("Unable to create the destination folder.\n")+
         AnsiString("Please review your entry for errors or\n")+
         AnsiString("illegal characters."),
         AppName+"Folder Creation Error",
         MB_OK
      );
      log("creating the app directory failed");
      return false;
   }
   bool ok=true;
   AnsiString appdir,appini,appdat,apppgp,fname,thedir,file;

   // open the uninstall dat file for file list
   AnsiString uninstlog=AddBackslash(AppDIR->Text)+UninstDat;
   TStringList *uninstdat = new TStringList();
   uninstdat->Sorted = true;
   uninstdat->Duplicates = dupIgnore;
   if (FileExists(uninstlog)) uninstdat->LoadFromFile(uninstlog);

   log("copying installation files to QS directory");
   TStringList *files = new TStringList();
   GetFilenames(files,AddBackslash(TempDir)+"*.*");
   for (int i=0; i<files->Count; i++)
   {
      file = files->Strings[i];

      DWORD attrib = GetFileAttributes(file.c_str());
      if ((attrib && FILE_ATTRIBUTE_DIRECTORY)==FILE_ATTRIBUTE_DIRECTORY)
         continue;

      fname = ExtractFileName(file); if (fname==SetupLog) continue;

      thedir = AddBackslash(AppDIR->Text);
      if (fname.LowerCase()=="setup.exe") fname = UnAppExe; // rename Setup to Uninst

      log("copying "+fname+" to "+thedir+fname);
      if ((ok=CopySetupFile(file,thedir+fname))==false)
      {
         log("copy failed");
         break;
      }
      uninstdat->Add(s(thedir+fname).LowerCase());//GetShortFileName(thedir+fname).LowerCase();
   }
   delete files;

   appdir = AddBackslash(AppDIR->Text);
   appini = appdir+AppINI;
   appdat = appdir+AppDAT;
   apppgp = ChangeFileExt(appdat,".pgp");
   if (ok && FileExists(appini)) // it will
   {
      if (!FileExists(appdat) && !FileExists(apppgp))
      {
         log("writing pristine database");
         if ((ok=CopySetupFile(appini.c_str(),appdat.c_str()))==false)
            log("writing database failed");
      }
   }
   // save file list for uninstall
   if (ok) uninstdat->SaveToFile(uninstlog);
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::WriteDesktopShortcut()
{
   if (!ShortBox->Checked) return true; // || !ShortBox->Enabled) return true;

   log("installing desktop shortcut");
   AnsiString TargetPath,LinkPath,IconPath,Description="QuickSilver";

   // it must be impossible to have not successfully created the application
   // directory, but this function is the home of an access violation
   if (!DirectoryExists(AppDIR->Text)) ForceDirectories(AppDIR->Text);
   if (!DirectoryExists(AppDIR->Text)) return true; // fail quitely

   TargetPath = s(AddBackslash(AppDIR->Text)+AppEXE).LowerCase();
   IconPath = TargetPath;

   // this directory is not neccessarily c:\windows
   LinkPath = GetWindowsDir()+"\\desktop";
   if (!DirectoryExists(LinkPath)) ForceDirectories(LinkPath);
   if (!DirectoryExists(LinkPath))
   {
      MsgBoxHwnd(Handle,
         s("Unable to locate the windows 'desktop' directory.\n")+
         s("I'll continue with the install, but without\n")+
         s("installing the desktop shortcut."),
         s("Problem"),
         MB_OK
      );
      return true;
   }
   LinkPath += "\\"+AppName+".LNK";

   CreateLink(TargetPath,LinkPath,IconPath,Description,"");
   return true;
}
//---------------------------------------------------------------------------
/*bool __fastcall TSetupForm::WriteDesktopShortcut()
{
   if (!ShortBox->Checked) return true; // || !ShortBox->Enabled) return true;

   log("installing desktop shortcut");
   AnsiString TargetPath,LinkPath,IconPath,Description;
   TargetPath = s(AddBackslash(AppDIR->Text)+AppEXE).LowerCase();
   LinkPath = "c:\\windows\\desktop\\"+AppName+".LNK";
   IconPath = TargetPath;
   Description = "QuickSilver";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");
   return true;
}
*/
//---------------------------------------------------------------------------
#define CSTR(s) AnsiString(s).c_str()
bool __fastcall TSetupForm::WriteProgGroup()
{
   if (!GroupBox->Checked) return true;// || !GroupBox->Enabled) return true;

   AnsiString winfolder = AddBackslash(GetWindowsDir());
   AnsiString appdir = AddBackslash(AppDIR->Text);
   AnsiString folder = winfolder+"start menu\\programs\\"+AppName;
   AnsiString appExe = appdir+AppEXE;
   AnsiString uniExe = appdir+UnAppExe;
   AnsiString setExe = appdir+"setup.exe";
   AnsiString hlpRtf = appdir+"QuickStart.rtf";
   AnsiString readMe = appdir+"ReadMe.txt";
   AnsiString toDo   = appdir+"ToDo.txt";
   AnsiString licTxt = appdir+"License.txt";
   AnsiString hisTxt = appdir+"History.txt";
   AnsiString TargetPath,LinkPath,IconPath,Description,Params;

   TRegIniFile *reg = new TRegIniFile("Software\\MicroSoft\\Windows\\CurrentVersion\\Explorer");
   folder = reg->ReadString("Shell Folders","Programs","").LowerCase();
   delete reg;

   folder += "\\"+AppName;
   if (!DirectoryExists(folder)) CreateDirectory(folder.c_str(),NULL);

   log("creating program group");
   TargetPath = licTxt.LowerCase();
   LinkPath = folder+"\\License.LNK";
   IconPath = TargetPath;
   Description = "License";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");

   TargetPath = appExe.LowerCase();
   LinkPath = folder+"\\"+AppName+".LNK";
   IconPath = TargetPath;
   Description = "QuickSilver";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");

   TargetPath = uniExe.LowerCase();
   LinkPath = folder+"\\Uninstall "+ChangeFileExt(AppExe,"")+".LNK";
   IconPath = TargetPath;
   Description = "Uninstall "+AppName;
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");

   TargetPath = hlpRtf.LowerCase();
   LinkPath = folder+"\\QuickStart.LNK";
   IconPath = TargetPath;
   Description = "Quick Start";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");

   TargetPath = readMe.LowerCase();
   LinkPath = folder+"\\ReadMe.LNK";
   IconPath = TargetPath;
   Description = "Read Me";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");

   if (!isUpdate)
      Exec95("explorer.exe "+folder,SW_SHOW);
   return true;
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::BackBtnClick(TObject *Sender)
{
   Notebook1->PageIndex--;
   BackBtn->Enabled = Notebook1->PageIndex > 0;
   NextBtn->Enabled = Notebook1->PageIndex != 1 || (!EAddr->Text.IsEmpty() && !Host->Text.IsEmpty());
   NextBtn->Caption = Notebook1->PageIndex==2 ? "Install":"Next";

   AnsiString appdat = AddBackslash(AppDIR->Text)+AppDAT;
   AnsiString apppgp = ChangeFileExt(appdat,".pgp");
   switch (Notebook1->PageIndex)
   {
      case 0:
         ActiveControl = AppDIR;
         break;
      case 1:
         ActiveControl = EAddr;
         break;
   }
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::NextBtnClick(TObject *Sender)
{
   if (NextBtn->Caption=="Install")
   {
      if (MsgBoxHwnd(Handle,"Install now?",AppName,MB_OKCANCEL)==IDCANCEL || !RunSetup())
         return;
   }
   int idx = Notebook1->PageIndex+1; if (idx>3) idx--;
   BackBtn->Enabled = idx > 0;
   NextBtn->Enabled = idx != 1 || (!EAddr->Text.IsEmpty() && !Host->Text.IsEmpty());
   NextBtn->Caption = idx==2 ? "Install":"Next";
   CancelBtn->Caption = idx==3 ? "Exit":"Cancel";
   switch (idx)
   {
      case 2:
         RichEdit3->Clear();
         RichEdit3->Lines->Add(AnsiString("Folder = ")+AppDIR->Text);
         RichEdit3->Lines->Add(AnsiString("Address = ")+EAddr->Text);
         RichEdit3->Lines->Add(AnsiString("MailHost = ")+Host->Text);
         if (!isUpdate)
         {
            RichEdit3->Lines->Add(AnsiString("ProgramGroup = ")+(GroupBox->Checked ? "yes":"no"));
            RichEdit3->Lines->Add(AnsiString("Desktop Shortcut = ")+(ShortBox->Checked ? "yes":"no"));
         }
         break;
      case 3:
         BackBtn->Enabled = false;
         NextBtn->Enabled = false;
         break;
   }
   Notebook1->PageIndex = idx;
   switch (Notebook1->PageIndex)
   {
      case 0: ActiveControl = AppDIR; break;
      case 1: ActiveControl = EAddr; break;
   }
   if (idx==3) MyBeep(MB_OK);
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::RunSetup()
{
   Cursor = crHourGlass;
   logdir = AddBackslash(AppDIR->Text);
   bool ok = DoWaitDlg(AppName,"...installing "+AppName.LowerCase()+"...",this) &&
      CopyFiles() && WriteDat() && WriteRegistry() && WriteDesktopShortcut() &&
         WriteProgGroup();
   log(ok ? "setup success":"setup failed");
   Cursor = crDefault;
   return ok;
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::EAddrChange(TObject *Sender)
{
   NextBtn->Enabled = !EAddr->Text.IsEmpty() && !Host->Text.IsEmpty();
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::HostChange(TObject *Sender)
{
   NextBtn->Enabled = !EAddr->Text.IsEmpty() && !Host->Text.IsEmpty();
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::BrowseBtnClick(TObject *Sender)
{
   AnsiString path = GetDirectoryPath(this);
   if (path!="cancel") AppDIR->Text = path;
}
//---------------------------------------------------------------------------
bool __fastcall TSetupForm::CopySetupFile(const AnsiString &ExistingName, const AnsiString &NewName,
   bool failifexists)
{
   bool ok = CopyFile(ExistingName.c_str(),NewName.c_str(),failifexists);
   if (!ok)
   {
      LPVOID lpMsgBuf;
      DWORD result = GetLastError();
      FormatMessage(
         FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,NULL,result,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&lpMsgBuf, 0, NULL
      );
      MsgBoxHwnd(Handle,
         s("System Error: "+AnsiString((char*)lpMsgBuf))+
         s("File: "+ExistingName),
        "Copy Failed",
         MB_OK
      );
   }
   return ok;
}
//---------------------------------------------------------------------------
void __fastcall TSetupForm::FormShow(TObject *Sender)
{
   if (isUpdate)
   {
      for (int i=1; i<=2; i++) NextBtn->Click();
      AnsiString appdat = AddBackslash(AppDIR->Text)+AppDAT;
      AnsiString apppgp = AddBackslash(AppDIR->Text)+ChangeFileExt(AppDAT,".pgp");
      if (!FileExists(appdat) && FileExists(apppgp))
      {
         RichEdit3->Lines->Text =
            s("I have found that your QS database is encrypted. Because of ")+
            s("this I am unable to retrieve current configuration values. ")+
            s("This is not a problem though. Click 'Install' to complete ")+
            s("program setup.");
         RichEdit4->Font->Color = clGray;
         BackBtn->Enabled = false;
      }
   }
}
//---------------------------------------------------------------------------
